/**
 * @file hash.c
 * @brief Implementacao basica de uma tabela de dispersao com
 *        resolucao por encadeamento linear.
 */
#include <stdlib.h>
#include <string.h>

#include "hash.h"

/* Lista de primos auxiliar, tem de ser aumentada se precisarmos de
 * guardar mais elementos. */
#define PRIMES 7
static int primes[PRIMES] = {16127, 34651, 68111, 131071, 260999, 524287, 999983};

/**
 * @brief Estrutura que representa a tabela de dispersao.
 */
struct SThash_s {
  int N, M, p;
  int (*hash)(const void *, int M);
  void ** tab;
};

/**
 * @brief Funcao que inicializa uma tabela de dispersao.
 *
 * @param 'hash' identifica a funcao de dispersao a utilizar nesta
 *        tabela de dispersao.
 * @return Uma tabela de dispersao.
 */
SThash
STinit(int (*hash)(const void *, int M))
{
  SThash st = malloc(sizeof(struct SThash_s));

  st->N = st->p = 0; 
  st->M = primes[st->p];
  st->hash = hash;

  st->tab = malloc(st->M*sizeof(void *));
  memset(st->tab, 0x0, st->M*sizeof(void *));

  return st;
}

/**
 * @brief Funcao que liberta a memoria associada a uma tabela de
 *        dipersao.
 *
 * @param 'h' identifica a tabela de dispersao a libertar.
 */
void
STfree(SThash h)
{
  free(h->tab);
  free(h);
}

/**
 * @brief Funcao para procurar numa tabela de dispersao.
 *
 * @para 'h' identifica a tabela onde procurar.
 * @param 'item' identifica o elemento a procurar.
 * @param 'cmp' identifica a funcao de comparacao a utilizar.
 *
 * @return O elemento se existir, NULL caso contrario.
 */
void *
STsearch(SThash h, void * item, int (*cmp)(const void *, const void *))
{ 
  int i = h->hash(item, h->M);

  while (h->tab[i] != NULL)
    if (cmp(item, h->tab[i]) == 0)
      return h->tab[i];
    else
      i = (i+1) % h->M;

  return NULL;
}

/**
 * @brief Funcao para apagar numa tabela de dispersao.
 *
 * @para 'h' identifica a tabela onde apagar.
 * @param 'item' identifica o elemento a apagar.
 * @param 'cmp' identifica a funcao de comparacao a utilizar.
 */
void
STdelete(SThash h, void * item, int (*cmp)(const void *, const void *))
{ 
  int j, i = h->hash(item, h->M);
  void * v;

  while (h->tab[i] != NULL)
    if (cmp(item, h->tab[i]) == 0)
      break;
    else
      i = (i+1) % h->M;

  if (h->tab[i] == NULL) return;

  h->tab[i] = NULL; 

  h->N--;
  for (j = (i+1) % h->M; h->tab[j] != NULL; j = (j+1) % h->M, h->N--) {
    v = h->tab[j]; 
    h->tab[j] = NULL;
    STinsert(h, v);
  }
}

/**
 * Funcao auxiliar que permitir aumentar o tamanho da tabela. Se nao
 * for possivel (ver lista de primos em cima), o programa termina de
 * imediato.
 */
static void
STexpand(SThash h) {
  int i; 
  void **t = h->tab;

  if (++(h->p) >= PRIMES)
    exit(1);
  
  h->N = 0;
  h->M = primes[h->p];
  h->tab = malloc(h->M*sizeof(void *));
  memset(h->tab, 0x0, h->M*sizeof(void *));
  for (i = 0; i < primes[h->p - 1]; i++)
    if (t[i] != NULL)
      STinsert(h, t[i]);
  free(t);
}

/**
 * @brief Funcao para inserir numa tabela de dispersao.
 *
 * @para 'h' identifica a tabela onde inserir.
 * @param 'item' identifica o elemento a inserir.
 */
void
STinsert(SThash h, void * item) {
  int i = h->hash(item, h->M);

  while (h->tab[i] != NULL)
    i = (i+1) % h->M;

  h->tab[i] = item;

  if (h->N++ > 0.6*h->M) 
    STexpand(h);
}

