/**
 * @file proj2.c
 * @brief Implementacao do segundo projecto de IAED:
 */
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "list.h"
#include "hash.h"

#define NPKG 5
#define MINC 128

/**
 * @brief Estrutura que representa uma chamada.
 */
struct call {
  unsigned long nr;
  char *name, pkg;
};

/**
 * Variaveis locais ao programa. A fila de prioridades e' representada
 * por pq e consiste em NPKG listas ligadas. Utilizamos duas tabelas
 * de dispersao, namehash e nrhash, respectivamente indexadas por nome
 * e numero, para acelerar a procura nas listas.
 */
static struct node pq[NPKG];
static SThash namehash;
static SThash nrhash;

/**
 * Funcao para libertar a memoria associada a uma chamada.
 */
static void
free_call(void * c) {
  free(((struct call *) c)->name);
  free(c);
}

/**
 * Funcao auxiliar para ler nomes (ilimitados).
 */
static char *
readname()
{
  size_t size = 0, i = 0;
  char * name = NULL, c;

  while ((c = getchar()) != ' ' && c != '\n') {
    /* Verificar se existe ainda espaco livre. */
    if (i >= size) {
      size += MINC;
      name = realloc(name, sizeof(char)*(size+1));
    }

    name[i++] = c;
  }
  name[i++] = '\0';
  /* name = realloc(name, i*sizeof(char)); */

  return name;
}

/**
 * Funcao que compara dois nos das listas com base no numero de
 * cliente associado.
 */
static int
nrcmp(const void *pa, const void *pb)
{
  struct call * la = (struct call *) ((link) pa)->value;
  struct call * lb = (struct call *) ((link) pb)->value;

  if (la->nr < lb->nr)
    return -1;
  if (la->nr > lb->nr)
    return 1;
  return 0;
}

/**
 * Funcao que compara dois nos das listas com base no nome de
 * cliente associado.
 */
static int
namecmp(const void *pa, const void *pb)
{
  struct call * la = (struct call *) ((link) pa)->value;
  struct call * lb = (struct call *) ((link) pb)->value;

  return strcmp(la->name, lb->name);
}

/**
 * Funcao de dispersao para os nos das listas com base no numero de
 * cliente associado.
 */
static int
nrhashfun(const void *p, int M)
{
  return ((struct call *) ((link) p)->value)->nr % M;
}

/**
 * Funcao de dispersao para os nos das listas com base no nome de
 * cliente associado.
 */
static int
namehashfun(const void *p, int M)
{
  char * v = ((struct call *) ((link) p)->value)->name;
  int h = 0, a = 127;

  for (; *v != '\0'; v++)
    h = (a*h + *v) % M;
  return h;
}

/**
 * @brief Funcao main.
 *
 * @return Codigo de estado de saida do programa.
 */
int
main(int argc, char * argv[])
{
  char cmd;
  struct call * c, ctmp;
  struct node ntmp;
  link ltmp;
  size_t i;

  /* Inicializacao das filas e das tabelas de dispersao. */
  memset(pq, 0x0, sizeof(struct node)*NPKG);
  for (i = 0; i < NPKG; i++)
    pq[i].next = pq[i].prev = &pq[i];
  namehash = STinit(namehashfun);
  nrhash = STinit(nrhashfun);

  /* Processar comando. */
  while ((cmd = getchar()) != 'x') {
    switch (cmd) {
      case 'a':
        /* Nova chamada. */
        c = malloc(sizeof(struct call));
        getchar(); /* catch ' ' */
        c->name = readname();
        scanf("%lu", &c->nr);
        scanf(" %c", &c->pkg);

        getchar(); /* catch '\n' */

        /* Inserir chamada na fila de espera. */
        list_insert(&pq[c->pkg - 'A'], c);

        /* Indexar o novo no adicionado 'a fila (lista duplamente
         * ligada). */
        STinsert(namehash, pq[c->pkg - 'A'].prev);
        STinsert(nrhash, pq[c->pkg - 'A'].prev);
        break;
      case 'u':
        /* Identificar cliente e guardar numa estrutura temporaria. */
        scanf("%lu", &ctmp.nr);
        scanf(" %c", &ctmp.pkg);

        getchar(); /* catch '\n' */

        /* Guardar chamada temporaria num no' tambem temporario. */
        ntmp.value = &ctmp;

        /* Procurar no' na tabela indexada por numero. */
        ltmp = (link) STsearch(nrhash, &ntmp, nrcmp);

        /* Verificar se existe de facto alteracao. */
        if (ctmp.pkg != ((struct call *) ltmp->value)->pkg) {

          /* Actualizar pacote. */
          c = (struct call *) ltmp->value;
          ((struct call *) ltmp->value)->pkg = ctmp.pkg;

          /* Mover no' entre as filas. */
          list_move(&pq[c->pkg - 'A'], ltmp);
        }
        break;
      case 'l':
        getchar(); /* catch '\n' */

        /* Procurar proxima chamda. */
        for (i = 0; i < NPKG && pq[i].next == &pq[i]; i++);

        /* Se existir, imprimir informacao. */
        if (i < NPKG) {
          c = (struct call *) pq[i].next->value;
          printf("%s %lu %c\n", c->name, c->nr, c->pkg);
        } else /* NA caso contrario. */
          printf("NA\n");
        break;
      case 'p':
        getchar(); /* catch '\n' */
        /* Procurar proxima chamada. */
        for (i = 0; i < NPKG && pq[i].next == &pq[i]; i++);

        /* Se existir, imprimir e apagar a chamada. */
        if (i < NPKG) {
          c = (struct call *) pq[i].next->value;
          printf("%s %lu %c\n", c->name, c->nr, c->pkg);

          /* Apagar em ambas as tabelas e na fila. */
          STdelete(namehash, pq[i].next, namecmp);
          STdelete(nrhash, pq[i].next, nrcmp);
          list_delete(pq[i].next);

          /* Libertar memoria associada 'a chamada. */
          free_call(c);
        } else /* NA caso contrario. */
          printf("NA\n");
        break;
      case 'r':
        getchar(); /* catch ' ' */

        /* Identificar cliente e guardar em estruturas temporarias. */
        ctmp.name = readname();
        ntmp.value = &ctmp;

        /* Procurar no' na tabela indexada por nome. */
        ltmp = (link) STsearch(namehash, &ntmp, namecmp);
        c = (struct call *) ltmp->value;

        /* Apagar em ambas as tabelas e na fila. */
        STdelete(namehash, ltmp, namecmp);
        STdelete(nrhash, ltmp, nrcmp);
        list_delete(ltmp);

        /* Libertar memoria associada 'a chamada e ao nome temporario. */
        free_call(c);
        free(ctmp.name);
    }
  }

  /* Libertar a memoria ainda alocada. */
  STfree(namehash);
  STfree(nrhash);
  for (i = 0; i < NPKG; i++)
    list_free(&pq[i], free_call);

  return EXIT_SUCCESS;
}

